@@ -10,7 +10,8 @@
 #ifndef LLD_ELF_FILESYSTEM_H
 #define LLD_ELF_FILESYSTEM_H
 
-#include "lld/Core/LLVM.h"
+#include "lld/Common/LLVM.h"
+#include <system_error>
 
 namespace lld {
 namespace elf {
